# Use Modern Python
from __future__ import unicode_literals, absolute_import, print_function

# System imports
import csv

# External libraries
import six

def parse_line(value, quotechar=None, **kwargs):
    """
    A simple wrapper to parse a single CSV value
    """

    quotechar = quotechar or '"'

    if six.PY2:
        quotechar = quotechar.encode('utf-8')

    for row in csv.reader([value], quotechar=quotechar, **kwargs):
        if six.PY3:
            return row # Already unicode in Python 3, hooray!
        else:
            return [s.decode('utf-8') for s in row] # Always binary in Python 2, fooey!


def line_to_string(value, **kwargs):
    """
    A simple wrapper to write one CSV line
    """

    fh = six.StringIO()
    csv.writer(fh, **kwargs).writerow(value)
    return fh.getvalue()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django_prbac.fields
from django.conf import settings
import jsonfield.fields
import django_prbac.models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Grant',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('assignment', jsonfield.fields.JSONField(default=dict, help_text='Assignment from parameters (strings) to values (any JSON-compatible value)', blank=True)),
            ],
            bases=(django_prbac.models.ValidatingModel, models.Model),
        ),
        migrations.CreateModel(
            name='Role',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('slug', models.CharField(help_text='The formal slug for this role, which should be unique', unique=True, max_length=256)),
                ('name', models.CharField(help_text='The friendly name for this role to present to users; this need not be unique.', max_length=256)),
                ('description', models.TextField(default='', help_text='A long-form description of the intended semantics of this role.', blank=True)),
                ('parameters', django_prbac.fields.StringSetField(default=set, help_text='A set of strings which are the parameters for this role. Entered as a JSON list.', blank=True)),
            ],
            bases=(django_prbac.models.ValidatingModel, models.Model),
        ),
        migrations.CreateModel(
            name='UserRole',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('role', models.OneToOneField(related_name='user_role', to='django_prbac.Role', on_delete=models.CASCADE)),
                ('user', models.OneToOneField(related_name='prbac_role', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
            ],
            bases=(django_prbac.models.ValidatingModel, models.Model),
        ),
        migrations.AddField(
            model_name='grant',
            name='from_role',
            field=models.ForeignKey(related_name='memberships_granted', to='django_prbac.Role', help_text='The sub-role begin granted membership or permission', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='grant',
            name='to_role',
            field=models.ForeignKey(related_name='members', to='django_prbac.Role', help_text='The super-role or permission being given', on_delete=models.CASCADE),
        ),
    ]



# Use modern Python
from __future__ import unicode_literals, absolute_import, print_function

# Django imports
from django.contrib import admin
from django import forms
try:
    from django.urls import reverse
except ImportError:
    # remove when supported django>=1.10
    from django.core.urlresolvers import reverse

# External librariess
import simplejson

# Local imports
import django_prbac.csv
from django_prbac.models import *
from django_prbac.forms import StringSetInput

__all__ = [
    'RoleAdmin',
    'RoleAdminForm',
    'GrantAdmin',
]

class RoleAdminForm(forms.ModelForm):
    class Meta:
        model = Role
        widgets = {
            'parameters': StringSetInput
        }
        exclude = []


class RoleAdmin(admin.ModelAdmin):

    model = Role
    form = RoleAdminForm

    def parameters__csv(self, instance):
        return django_prbac.csv.line_to_string(sorted(list(instance.parameters)))
    parameters__csv.short_description = 'Parameters'
    parameters__csv.admin_order_field = 'parameters'

    list_display = [
        'slug',
        'name',
        'parameters__csv',
        'description',
    ]

    search_fields = [
        'slug',
        'name',
        'parameters',
        'description',
    ]

class GrantAdmin(admin.ModelAdmin):

    model = Grant

    def edit_link(self, instance):
        if instance.id:
            return '<a href="%s">(edit)</a>' % reverse('admin:django_prbac_grant_change',
                                                                         args=[instance.id])
        else:
            return ''
    edit_link.allow_tags = True
    edit_link.short_description = ''

    def assignment__dumps(self, instance):
        return simplejson.dumps(instance.assignment)
    assignment__dumps.short_description = 'Assignment'

    list_display = [
        'edit_link',
        'from_role',
        'to_role',
        'assignment__dumps',
    ]

    search_fields = [
        'from_role__name',
        'from_role__description',
        'to_role__name',
        'to_role__description',
    ]

    def get_queryset(self, request):
        return Grant.objects.select_related('to_role', 'from_role')

admin.site.register(Role, RoleAdmin)
admin.site.register(Grant, GrantAdmin)




# Use modern Python
from __future__ import unicode_literals, absolute_import, print_function

# Standard Library Imports
import time
import weakref

# Django imports
from django.db import models
from django.conf import settings

# External Library imports
import jsonfield

# Local imports
from django_prbac.fields import StringListField, StringSetField


__all__ = [
    'Role',
    'Grant',
    'RoleInstance',
    'UserRole',
]

class ValidatingModel(object):
    def save(self, force_insert=False, force_update=False, **kwargs):
        if not (force_insert or force_update):
            self.full_clean() # Will raise ValidationError if needed
        super(ValidatingModel, self).save(force_insert, force_update, **kwargs)


class Role(ValidatingModel, models.Model):
    """
    A PRBAC role, aka a Role parameterized by a set of named variables. Roles
    also model privileges:  They differ only in that privileges only refer
    to real-world consequences when all parameters are instantiated.
    """

    PRIVILEGES_BY_SLUG = "DJANGO_PRBAC_PRIVELEGES"
    ROLES_BY_ID = "DJANGO_PRBAC_ROLES"
    _default_instance = lambda s:None


    # Databaes fields
    # ---------------

    slug = models.CharField(
        max_length=256,
        help_text='The formal slug for this role, which should be unique',
        unique=True,
    )

    name = models.CharField(
        max_length=256,
        help_text='The friendly name for this role to present to users; this need not be unique.',
    )

    description = models.TextField(
        help_text='A long-form description of the intended semantics of this role.',
        blank=True,
        default='',
    )

    parameters = StringSetField(
        help_text='A set of strings which are the parameters for this role. Entered as a JSON list.',
        blank=True,
        default=set,
    )

    class Meta:
        app_label = 'django_prbac'

    # Methods
    # -------

    @classmethod
    def get_cache(cls):
        try:
            cache = cls.cache
        except AttributeError:
            timeout = getattr(settings, 'DJANGO_PRBAC_CACHE_TIMEOUT', 60)
            cache = cls.cache = DictCache(timeout)
        return cache

    @classmethod
    def update_cache(cls):
        roles = cls.objects.prefetch_related('memberships_granted').all()
        roles = {role.id: role for role in roles}
        for role in roles.values():
            role._granted_privileges = privileges = []
            # Prevent extra queries by manually linking grants and roles
            # because Django 1.6 isn't smart enough to do this for us
            for membership in role.memberships_granted.all():
                membership.to_role = roles[membership.to_role_id]
                membership.from_role = roles[membership.from_role_id]
                privileges.append(membership.instantiated_to_role({}))
        cache = cls.get_cache()
        cache.set(cls.ROLES_BY_ID, roles)
        cache.set(cls.PRIVILEGES_BY_SLUG,
            {role.slug: role.instantiate({}) for role in roles.values()})

    @classmethod
    def get_privilege(cls, slug, assignment=None):
        """
        Optimized lookup of privilege by slug

        This optimization is specifically geared toward cases where
        `assignments` is empty.
        """
        cache = cls.get_cache()
        if cache.disabled:
            roles = Role.objects.filter(slug=slug)
            if roles:
                return roles[0].instantiate(assignment or {})
            return None
        privileges = cache.get(cls.PRIVILEGES_BY_SLUG)
        if privileges is None:
            cls.update_cache()
            privileges = cache.get(cls.PRIVILEGES_BY_SLUG)
        privilege = privileges.get(slug)
        if privilege is None:
            return None
        if assignment:
            return privilege.role.instantiate(assignment)
        return privilege

    def get_cached_role(self):
        """
        Optimized lookup of role by id
        """
        cache = self.get_cache()
        if cache.disabled:
            return self
        roles = cache.get(self.ROLES_BY_ID)
        if roles is None or self.id not in roles:
            self.update_cache()
            roles = cache.get(self.ROLES_BY_ID)
        return roles.get(self.id, self)

    def get_privileges(self, assignment):
        if not assignment:
            try:
                return self._granted_privileges
            except AttributeError:
                pass
        return [membership.instantiated_to_role(assignment)
                for membership in self.memberships_granted.all()]

    def instantiate(self, assignment):
        """
        An instantiation of this role with some parameters fixed via the provided assignments.
        """
        if assignment:
            filtered_assignment = {key: assignment[key]
                for key in self.parameters & set(assignment.keys())}
        else:
            value = self._default_instance()
            if value is not None:
                return value
            filtered_assignment = assignment
        value = RoleInstance(self, filtered_assignment)
        if not filtered_assignment:
            self._default_instance = weakref.ref(value)
        return value

    def has_privilege(self, privilege):
        """
        Shortcut for checking privileges easily for roles with no params (aka probably users)
        """
        role = self.get_cached_role()
        return role.instantiate({}).has_privilege(privilege)

    @property
    def assignment(self):
        """
        A Role stored in the database always has an empty assignment.
        """
        return {}

    def __repr__(self):
        return 'Role(%r, parameters=%r)' % (self.slug, self.parameters)

    def __unicode__(self):
        return '%s (%s)' % (self.name, self.slug)


class Grant(ValidatingModel, models.Model):
    """
    A parameterized membership between a sub-role and super-role.
    The parameters applied to the super-role are all those.
    """


    # Database Fields
    # ---------------

    from_role = models.ForeignKey(
        'Role',
        help_text='The sub-role begin granted membership or permission',
        related_name='memberships_granted',
        on_delete=models.CASCADE,
    )

    to_role = models.ForeignKey(
        'Role',
        help_text='The super-role or permission being given',
        related_name='members',
        on_delete=models.CASCADE,
    )

    assignment = jsonfield.JSONField(
        help_text='Assignment from parameters (strings) to values (any JSON-compatible value)',
        blank=True,
        default=dict,
    )

    class Meta:
        app_label = 'django_prbac'

    # Methods
    # -------

    def instantiated_to_role(self, assignment):
        """
        Returns the super-role instantiated with the parameters of the membership
        composed with the `parameters` passed in.
        """
        composed_assignment = {}
        if assignment:
            for key in self.to_role.parameters & set(assignment.keys()):
                composed_assignment[key] = assignment[key]
        if self.assignment:
            composed_assignment.update(self.assignment)
        return self.to_role.instantiate(composed_assignment)

    def __repr__(self):
        return 'Grant(from_role=%r, to_role=%r, assignment=%r)' % (self.from_role, self.to_role, self.assignment)


class UserRole(ValidatingModel, models.Model):
    """
    A link between a django.contrib.auth.models.User and
    a django_prbac.models.Role. They are kept to
    one-to-one fields to make their use extremely simple:

    request.user.prbac_role.has_privilege(...)
    """

    user = models.OneToOneField(settings.AUTH_USER_MODEL, related_name='prbac_role', on_delete=models.CASCADE)
    role = models.OneToOneField(Role, related_name='user_role', on_delete=models.CASCADE)

    class Meta:
        app_label = 'django_prbac'

    def has_privilege(self, privilege):
        return self.role.has_privilege(privilege)

    def __eq__(self, other):
        return self.user == other.user and self.role == other.role

    def __repr__(self):
        return 'UserRole(user=%r, role=%r)' % (self.user, self.role)


class RoleInstance(object):
    """
    A parameterized role along with some parameters that are fixed. Note that this is
    not a model but only a transient Python object.
    """

    def __init__(self, role, assignment):
        self.role = role
        self.assignment = assignment
        self.slug = role.slug
        self.name = role.name
        self.parameters = role.parameters - set(assignment.keys())

    def instantiate(self, assignment):
        """
        This role further instantiated with the additional assignment.
        Note that any parameters that are already fixed are not actually
        available for being assigned, so will _not_ change.
        """
        composed_assignment = {}
        if assignment:
            for key in self.parameters & set(assignment.keys()):
                composed_assignment[key] = assignment[key]
        if self.assignment:
            composed_assignment.update(self.assignment)
        # this seems like a bug (wrong arguments). is this method ever called?
        return RoleInstance(composed_assignment)

    def has_privilege(self, privilege):
        """
        True if this instantiated role is allowed the privilege passed in,
        (which is itself an RoleInstance)
        """

        if self == privilege:
            return True

        return any(p.has_privilege(privilege)
                   for p in self.role.get_privileges(self.assignment))

    def __eq__(self, other):
        return self.slug == other.slug and self.assignment == other.assignment

    def __repr__(self):
        return 'RoleInstance(%r, parameters=%r, assignment=%r)' % (self.slug, self.parameters, self.assignment)


class DictCache(object):
    """A simple in-memory dict cache

    :param timeout: Number of seconds until an item in the cache expires.
    """

    def __init__(self, timeout=60):
        self.timeout = timeout
        self.data = {}

    @property
    def disabled(self):
        return self.timeout == 0

    def get(self, key, default=None):
        now = time.time()
        value, expires = self.data.get(key, (default, now))
        if now > expires:
            self.data.pop(key)
            return default
        return value

    def set(self, key, value):
        self.data[key] = (value, time.time() + self.timeout)

    def clear(self):
        self.data.clear()



# Use Modern Python
from __future__ import unicode_literals, absolute_import, print_function

# Django imports
from django.test import TestCase
from django.http import HttpRequest

# Local imports
from django_prbac.decorators import requires_privilege
from django_prbac.exceptions import PermissionDenied
from django_prbac.models import Role
from django_prbac import arbitrary

class TestDecorators(TestCase):

    def setUp(self):
        Role.get_cache().clear()
        self.zazzle_privilege = arbitrary.role(slug=arbitrary.unique_slug('zazzle'), parameters=set(['domain']))

    def test_requires_privilege_no_current_role(self):
        """
        When a privilege is required but there is no role attached
        to the current request, permission is denied. No crashing.
        """
        @requires_privilege(self.zazzle_privilege.slug, domain='zizzle')
        def view(request, *args, **kwargs):
            pass

        request = HttpRequest()
        with self.assertRaises(PermissionDenied):
            view(request)

    def test_requires_privilege_no_such(self):
        """
        When a required privilege is not even defined in the database,
        permission is denied; no crashing.
        """
        @requires_privilege('bomboozle', domain='zizzle')
        def view(request, *args, **kwargs):
            pass


        requestor_role = arbitrary.role()
        request = HttpRequest()
        request.role = requestor_role
        with self.assertRaises(PermissionDenied):
            view(request)

    def test_requires_privilege_denied(self):
        """
        When a privilege exists but the current
        role does not have access to it, permission
        is denied
        """

        @requires_privilege(self.zazzle_privilege.slug, domain='zizzle')
        def view(request, *args, **kwargs):
            pass

        requestor_role = arbitrary.role()

        request = HttpRequest()
        request.role = requestor_role.instantiate({})
        with self.assertRaises(PermissionDenied):
            view(request)


    def test_requires_privilege_wrong_param(self):

       @requires_privilege(self.zazzle_privilege.slug, domain='zizzle')
       def view(request, *args, **kwargs):
           pass

       requestor_role = arbitrary.role()
       arbitrary.grant(from_role=requestor_role, to_role=self.zazzle_privilege, assignment=dict(domain='whapwhap'))

       request = HttpRequest()
       request.role = requestor_role.instantiate({})
       with self.assertRaises(PermissionDenied):
           view(request)

    def test_requires_privilege_ok(self):

        @requires_privilege(self.zazzle_privilege.slug, domain='zizzle')
        def view(request, *args, **kwargs):
            pass

        requestor_role = arbitrary.role()
        arbitrary.grant(from_role=requestor_role, to_role=self.zazzle_privilege, assignment=dict(domain='zizzle'))

        request = HttpRequest()
        request.role = requestor_role.instantiate({})
        view(request)

    def test_requires_privilege_role_on_user_ok(self):
        """
        Verify that privilege is recognized when the request user has the prbac_role, but request.role is not set.
        """

        @requires_privilege(self.zazzle_privilege.slug, domain='zizzle')
        def view(request, *args, **kwargs):
            pass

        user = arbitrary.user()
        requestor_role = arbitrary.role()
        arbitrary.grant(from_role=requestor_role, to_role=self.zazzle_privilege, assignment=dict(domain='zizzle'))
        arbitrary.user_role(user=user, role=requestor_role)

        request = HttpRequest()
        request.user = user
        view(request)



# Use Modern Python
from __future__ import unicode_literals, absolute_import, print_function

# Django imports
from django.test import TestCase

# Local imports
from django_prbac.fields import StringListField, StringSetField

class TestStringListField(TestCase):
    """
    Test suite for django_prbac.fields.StringListField
    """

    def test_is_string_list(self):
        field = StringListField('testing')

        self.assertTrue(field.is_string_list([]))
        self.assertTrue(field.is_string_list(["hello", "goodbye"]))

        self.assertFalse(field.is_string_list("boo"))
        self.assertFalse(field.is_string_list(3))
        self.assertFalse(field.is_string_list('"A","B"'))


    def test_to_python_convert(self):
        field = StringListField('testing')
        self.assertEqual(field.to_python(''), [])
        self.assertEqual(field.to_python('"A","B","C"'), ['A', 'B', 'C'])


    def test_to_python_already_done(self):
        field = StringListField('testing')
        self.assertEqual(field.to_python([]), [])
        self.assertEqual(field.to_python(["A","B","C"]), ['A', 'B', 'C'])

        with self.assertRaises(ValueError):
            field.to_python(4)

        with self.assertRaises(ValueError):
            field.to_python([1, 2, 3])

        with self.assertRaises(ValueError):
            field.to_python(None)


    def test_get_prep_value_convert(self):
        field = StringListField('testing')

        self.assertEqual(field.get_prep_value(["A", "B", "C"]), 'A,B,C')
        self.assertEqual(field.get_prep_value(["A", "B,C", "D"]), 'A,"B,C",D')

        with self.assertRaises(ValueError):
            field.get_prep_value(5)


class TestStringSetField(TestCase):
    """
    Test suite for django_prbac.fields.StringSetField
    """

    def test_is_string_set(self):
        field = StringSetField('testing')

        self.assertTrue(field.is_string_set(set([])))
        self.assertTrue(field.is_string_set(set(["hello", "goodbye"])))

        self.assertFalse(field.is_string_set(["A", "B"]))
        self.assertFalse(field.is_string_set("boo"))
        self.assertFalse(field.is_string_set(3))
        self.assertFalse(field.is_string_set('["A", "B"]'))

    def test_to_python_convert(self):
        field = StringSetField('testing')

        # This that are legitimate to store in the DB
        self.assertEqual(field.to_python(''), set())
        self.assertEqual(field.to_python('"A","B","C"'), set(['A', 'B', 'C']))


    def test_to_python_already_done(self):
        field = StringSetField('testing')
        self.assertEqual(field.to_python([]), set())
        self.assertEqual(field.to_python(set(["A","B","C"])), set(['A', 'B', 'C']))

        with self.assertRaises(ValueError):
            field.to_python(4)

        with self.assertRaises(ValueError):
            field.to_python([1, 2, 3])

        with self.assertRaises(ValueError):
            field.to_python(None)


    def test_get_prep_value_convert(self):
        field = StringSetField('testing')

        self.assertEqual(field.get_prep_value(set(["A", "B", "C"])), 'A,B,C')
        self.assertEqual(field.get_prep_value(set(["C", "B", "A"])), 'A,B,C')

        with self.assertRaises(ValueError):
            field.get_prep_value(5)





# Use modern Python
from __future__ import unicode_literals, absolute_import, print_function

# Standard Library Imports

# Django imports
from django.test import TestCase # https://code.djangoproject.com/ticket/20913

# External Library imports

# Local imports
from django_prbac.models import *
from django_prbac import arbitrary


class TestRole(TestCase):

    def setUp(self):
        Role.get_cache().clear()

    def test_has_permission_immediate_no_params(self):
        subrole = arbitrary.role()
        superrole1 = arbitrary.role()
        superrole2 = arbitrary.role()
        arbitrary.grant(to_role=superrole1, from_role=subrole)

        # A few ways of saying the same thing
        self.assertTrue(subrole.instantiate({}).has_privilege(superrole1.instantiate({})))
        self.assertTrue(subrole.has_privilege(superrole1.instantiate({})))
        self.assertTrue(subrole.instantiate({}).has_privilege(superrole1))
        self.assertTrue(subrole.has_privilege(superrole1))

        self.assertFalse(subrole.instantiate({}).has_privilege(superrole2.instantiate({})))
        self.assertFalse(subrole.has_privilege(superrole2.instantiate({})))
        self.assertFalse(subrole.instantiate({}).has_privilege(superrole2))
        self.assertFalse(subrole.has_privilege(superrole2))


    def test_has_permission_transitive_no_params(self):
        subrole = arbitrary.role()
        midrole = arbitrary.role()
        superrole1 = arbitrary.role()
        superrole2 = arbitrary.role()
        arbitrary.grant(to_role=midrole, from_role=subrole)
        arbitrary.grant(to_role=superrole1, from_role=midrole)

        # A few ways of saying the same thing
        self.assertTrue(subrole.instantiate({}).has_privilege(superrole1.instantiate({})))
        self.assertTrue(subrole.has_privilege(superrole1.instantiate({})))
        self.assertTrue(subrole.instantiate({}).has_privilege(superrole1))
        self.assertTrue(subrole.has_privilege(superrole1))

        self.assertFalse(subrole.instantiate({}).has_privilege(superrole2.instantiate({})))
        self.assertFalse(subrole.has_privilege(superrole2.instantiate({})))
        self.assertFalse(subrole.instantiate({}).has_privilege(superrole2))
        self.assertFalse(subrole.has_privilege(superrole2))


    def test_has_permission_far_transitive_no_params(self):
        subrole = arbitrary.role()
        superrole1 = arbitrary.role()
        superrole2 = arbitrary.role()

        midroles = [arbitrary.role() for __ in range(0, 10)]

        arbitrary.grant(subrole, midroles[0])
        arbitrary.grant(midroles[-1], superrole1)

        # Link up all roles in the list that are adjacent
        for midsubrole, midsuperrole in zip(midroles[:-1], midroles[1:]):
            arbitrary.grant(from_role=midsubrole, to_role=midsuperrole)

        self.assertTrue(subrole.instantiate({}).has_privilege(superrole1.instantiate({})))
        self.assertFalse(subrole.instantiate({}).has_privilege(superrole2.instantiate({})))


    def test_has_permission_immediate_params(self):
        subrole = arbitrary.role()
        superrole1 = arbitrary.role(parameters=set(['one']))
        arbitrary.grant(to_role=superrole1, from_role=subrole, assignment=dict(one='foo'))

        self.assertTrue(subrole.instantiate({}).has_privilege(superrole1.instantiate(dict(one='foo'))))
        self.assertFalse(subrole.instantiate({}).has_privilege(superrole1.instantiate(dict(one='baz'))))


    def test_unsaved_role_does_not_have_permission(self):
        role1 = Role()
        role2 = arbitrary.role()
        self.assertFalse(role1.has_privilege(role2))
        self.assertFalse(role2.has_privilege(role1))


class TestGrant(TestCase):

    def test_instantiated_to_role_smoke_test(self):
        """
        Basic smoke test:
        1. grant.instantiated_role({})[param] == grant.assignment[param] if param is free for the role
        2. grant.instantiated_role({})[param] does not exist if param is not free for the role
        """

        parameters = ['one']

        superrole = arbitrary.role(parameters=parameters)
        grant = arbitrary.grant(to_role=superrole, assignment={'one':'hello'})
        self.assertEqual(grant.instantiated_to_role({}).assignment, {'one':'hello'})

        grant = arbitrary.grant(to_role=superrole, assignment={'two': 'goodbye'})
        self.assertEqual(grant.instantiated_to_role({}).assignment, {})


class TestUserRole(TestCase):

    def setUp(self):
        Role.get_cache().clear()

    def test_user_role_integration(self):
        """
        Basic smoke test of integration of PRBAC with django.contrib.auth
        """
        user = arbitrary.user()
        role = arbitrary.role()
        priv = arbitrary.role()
        arbitrary.grant(from_role=role, to_role=priv)
        user_role = arbitrary.user_role(user=user, role=role)

        self.assertEqual(user.prbac_role, user_role)
        self.assertTrue(user.prbac_role.has_privilege(role))
        self.assertTrue(user.prbac_role.has_privilege(priv))




"""
Settings just for running the django-prbac tests or checking out
the admin site.
"""

SECRET_KEY='Not a secret key at all, actually'

DEBUG=True

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'django-prbac.db',
    }
}

INSTALLED_APPS = [
    # Django apps necessary to run the admin site
    'django.contrib.auth',
    'django.contrib.admin',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.staticfiles',

    # And this app
    'django_prbac',
]

STATIC_URL='/static/'

ROOT_URLCONF='django_prbac.urls'



# Use Modern Python
from __future__ import unicode_literals, absolute_import, print_function

# Django imports
from django.test import TestCase

# Local imports
from django_prbac.forms import StringListFormField

class TestStringListFormField(TestCase):
    """
    Test suite for django_prbac.fields.StringListField
    """

    def test_clean(self):
        field = StringListFormField(required=False, skipinitialspace=False)

        self.assertEqual(field.clean('hello, goodbye'), ['hello', ' goodbye'])
        self.assertEqual(field.clean('hello,goodbye'), ['hello', 'goodbye'])
        self.assertEqual(field.clean('"hello",    goodbye'), ['hello', '    goodbye'])
        self.assertEqual(field.clean('"hello"," oh no "'), ['hello', ' oh no '])
        self.assertEqual(field.clean('"hello","one,two"'), ['hello', 'one,two'])

    def test_quotechar(self):
        field = StringListFormField(required=False, quotechar='|')

        self.assertEqual(field.clean('hello, goodbye'), ['hello', 'goodbye'])
        self.assertEqual(field.clean('hello,goodbye'), ['hello', 'goodbye'])
        self.assertEqual(field.clean('hello,    goodbye'), ['hello', 'goodbye'])
        self.assertEqual(field.clean('"hello",    goodbye'), ['"hello"', 'goodbye'])
        self.assertEqual(field.clean('"hello","oh, no"'), ['"hello"', '"oh', 'no"'])
        self.assertEqual(field.clean('hello,|oh, no|'), ['hello', 'oh, no'])



# Use Modern Python
from __future__ import unicode_literals, absolute_import, print_function

# System imports

# Django imports
from django.forms import ValidationError, CharField, TextInput

# External libraries
import six

# Local imports
import django_prbac.csv


class StringListInput(TextInput):
    def render(self, name, value, attrs=None):
        print('poops', name, value)
        if isinstance(value, six.string_types):
            return super(StringListInput, self).render(name, value)
        else:
            rendered_value = django_prbac.csv.line_to_string(list(value))
            return super(StringListInput, self).render(name, rendered_value)


class StringSetInput(TextInput):
    def render(self, name, value, attrs=None):
        print('poops', name, value)
        if isinstance(value, six.string_types):
            return super(StringSetInput, self).render(name, value)
        else:
            rendered_value = django_prbac.csv.line_to_string(sorted(list(value)))
            return super(StringSetInput, self).render(name, rendered_value)


class StringListFormField(CharField):
    """
    A Django form field for lists of strings separated by commas, quotes optional
    """
    def __init__(self, quotechar=None, skipinitialspace=None, *args, **kwargs):
        self.quotechar = (quotechar or '"')
        self.skipinitialspace = True if skipinitialspace is None else skipinitialspace
        defaults = {'widget': StringListInput}
        defaults.update(kwargs)
        super(StringListFormField, self).__init__(*args, **defaults)

    def is_string_list(self, value):
        return isinstance(value, list) and all([isinstance(v, six.string_types) for v in value])

    def clean(self, value):
        if self.is_string_list(value):
            return value

        elif not isinstance(value, six.string_types):
            raise ValidationError('%r cannot be converted to a string list' % value)

        else:
            try:
                return django_prbac.csv.parse_line(
                    value,
                    skipinitialspace=self.skipinitialspace,
                    quotechar=self.quotechar,
                )

            except ValueError:
                raise ValidationError('%r cannot be converted to a string list' % value)




# Use modern Python
from __future__ import unicode_literals, absolute_import, print_function

from functools import wraps

# Local Imports
from django.http import Http404
from django_prbac.exceptions import PermissionDenied
from django_prbac.utils import has_privilege


def requires_privilege(slug, **assignment):
    def decorate(fn):
        """
        Returns a function equivalent to `fn` but that requires
        a role with slug `slug` to be reachable from `request.role`
        or `request.user.prbac_role`
        with the parameters specified in `assignment`
        (in a parameterized fashion)
        """
        @wraps(fn)
        def wrapped(request, *args, **kwargs):
            if not has_privilege(request, slug, **assignment):
                raise PermissionDenied()
            return fn(request, *args, **kwargs)

        return wrapped

    return decorate


def requires_privilege_raise404(slug, **assignment):
    """
    A version of the requires_privilege decorator which raises an Http404
    if PermissionDenied is raised.
    """
    def decorate(fn):
        @wraps(fn)
        def wrapped(request, *args, **kwargs):
            if not has_privilege(request, slug, **assignment):
                raise Http404()
            return fn(request, *args, **kwargs)
        return wrapped
    return decorate



from setuptools import setup, find_packages

setup(
    name='django-prbac',
    version='0.0.4',
    description='Parameterized Role-Based Access Control for Django',
    author='Dimagi',
    author_email='dev@dimagi.com',
    url='http://github.com/dimagi/django-prbac',
    packages=find_packages(),
    zip_safe=False,
    install_requires=[
        'django>=1.8,<1.11',
        'jsonfield>=1.0.3',
        'simplejson',
        'six',
    ],
)



# Use Modern Python
from __future__ import unicode_literals, absolute_import, print_function

# Django imports
from django.db import models

# External libraries
import six

# Local imports
import django_prbac.csv
from django_prbac.forms import StringListFormField


class StringListField(models.TextField):
    """
    A Django field for lists of strings
    """

    def is_string_list(self, value):
        return isinstance(value, list) and all([isinstance(v, six.string_types) for v in value])

    def to_python(self, value):
        """
        Best-effort conversion of "any value" to a string list.

        It does not try that hard, because curious values probably indicate
        a mistake and we should fail early.
        """

        # Already the appropriate python type
        if self.is_string_list(value):
            return value
        elif isinstance(value, six.string_types):
            return django_prbac.csv.parse_line(value)
        else:
            raise ValueError('Invalid value for StringListField: %r is neither the correct type nor deserializable' % value)

    def get_prep_value(self, value):
        """
        Converts the value, which must be a string list, to a comma-separated string,
        quoted appropriately. This format is private to the field type so it is not
        exposed for customization or any such thing.
        """

        if not self.is_string_list(value):
            raise ValueError('Invalid value for StringListField: %r' % value)
        else:
            return django_prbac.csv.line_to_string(value, lineterminator='')

    def value_to_string(self, obj):
        value = self._get_val_from_obj(obj)
        return self.get_prep_value(value)

    def formfield(self, **kwargs):
        """
        The default form field is a StringListFormField.
        """

        defaults = {'form_class': StringListFormField}
        defaults.update(kwargs)
        return super(StringListField, self).formfield(**defaults)

    def from_db_value(self, value, expression, connection, context):
        return self.to_python(value)


class StringSetField(StringListField):
    """
    A Django field for set of strings.
    """

    # TODO thought: If Python had polymorphism this ought be "Serialize a => Field (List a)"

    def is_string_set(self, value):
        return isinstance(value, set) and all([isinstance(v, six.string_types) for v in value])

    def to_python(self, value):
        """
        Best-effort conversion of "any value" to a string set. Mostly strict,
        but a bit lenient to allow lists to be passed in by form fields.
        """

        # Already the appropriate python type
        if self.is_string_set(value):
            return value

        # If it is a string list, we will turn it into a set; this lenience let's us
        # re-use the form field easily
        if self.is_string_list(value):
            return set(value)

        # First let StringListField do whatever it needs to do; this will now be a string list
        try:
            value = super(StringSetField, self).to_python(value)
        except ValueError:
            raise ValueError('Invalid value for StringSetField: %r' % value)

        return set(value)

    def value_to_string(self, obj):
        value = self._get_val_from_obj(obj)
        return self.get_prep_value(value)

    def get_prep_value(self, value):
        if not self.is_string_set(value):
            raise ValueError('Invalid value %r for StringSetField' % value)
        else:
            return super(StringSetField, self).get_prep_value(sorted(value))

    def from_db_value(self, value, expression, connection, context):
        return self.to_python(value)



# Use modern Python
from __future__ import unicode_literals, absolute_import, print_function
import warnings

# Local Imports
from django_prbac.exceptions import PermissionDenied
from django_prbac.models import Role, UserRole


def has_privilege(request, slug, **assignment):
    """
    Returns true if the request has the privilege specified by slug,
    otherwise false
    """
    privilege = Role.get_privilege(slug, assignment)
    if privilege is None:
        return False

    if hasattr(request, 'role'):
        if request.role.has_privilege(privilege):
            return True

    if hasattr(request, 'user') and hasattr(request.user, 'prbac_role'):
        try:
            request.user.prbac_role
        except UserRole.DoesNotExist:
            return False
        return request.user.prbac_role.has_privilege(privilege)

    return False


def ensure_request_has_privilege(request, slug, **assignment):
    """
    DEPRECATED
    """
    warnings.warn(
        '`ensure_request_has_privilege` is deprecated. You likely want '
        '`has_permission` or one of the `requires_privilege` decorators',
        DeprecationWarning
    )
    if not has_privilege(request, slug, **assignment):
        raise PermissionDenied()



# Use modern Python
from __future__ import unicode_literals, absolute_import, print_function

# Standard Library Imports
from random import choice
import uuid

# Django imports
from django.contrib.auth.models import User

# Local Imports
from django_prbac.models import *

__all__ = [
    'role',
    'grant',
    'unique_name',
]

def instantiate(generator_or_value):
    """
    Dynamic typing hack to try to call generators if provided,
    otherwise return the value directly if not callable. This will
    break badly if used for values that can be callable.
    """

    if callable(generator_or_value):
        return generator_or_value()
    else:
        return generator_or_value

def arbitrary_slug():
    return choice(['foo', 'bar', 'baz', 'zizzle', 'zazzle'])


def arbitrary_unique_slug(prefix=None, suffix=None):
    prefix = instantiate(prefix or '')
    suffix = instantiate(suffix or '')
    return prefix + arbitrary_slug() + uuid.uuid4().hex + suffix


def arbitrary_user(username=None, password=None, email=None, save=True, **kwargs):
    username = instantiate(username or arbitrary_unique_slug)[:74]
    password = instantiate(password or arbitrary_unique_slug)[:74]
    email = instantiate(email) if email is not None else ('%s@%s.com' % (arbitrary_unique_slug(), arbitrary_unique_slug()))[:74]

    user = User(username=username,
                password=password,
                email=email,
                **kwargs)

    if save:
        user.save()

    return user


def arbitrary_role(slug=None, name=None, save=True, **kwargs):
    slug = instantiate(slug or arbitrary_unique_slug)
    name = instantiate(name or arbitrary_slug)

    role = Role(
        slug=slug,
        name=name,
        **kwargs
    )

    if save:
        role.save()

    return role


def arbitrary_grant(from_role=None, to_role=None, save=True, **kwargs):
    from_role = instantiate(from_role if from_role is not None else arbitrary_role)
    to_role = instantiate(to_role if to_role is not None else arbitrary_role)

    grant = Grant(
        from_role=from_role,
        to_role=to_role,
        **kwargs
    )

    if save:
        grant.save()

    return grant

def arbitrary_user_role(user=None, role=None, save=True, **kwargs):
    user = instantiate(user or arbitrary_user)
    role = instantiate(role or arbitrary_role)

    user_role = UserRole(user=user,
                         role=role,
                         **kwargs)

    if save:
        user_role.save()

    return user_role

role = arbitrary_role
grant = arbitrary_grant
unique_slug = arbitrary_unique_slug
user = arbitrary_user
user_role = arbitrary_user_role
